import logging
import time
import re

from avocado.utils import process
from virttest.env_process import preprocess
from virttest.virt_vm import VMDeadKernelCrashError
from virttest import error_context
from virttest import utils_test
from virttest import utils_time


@error_context.context_aware
def run(test, params, env):
    """
    Time clock offset check when guest crash/bsod test:

    1) boot guest with '-rtc base=utc,clock=host,driftfix=slew';
    2) sync host system time with ntp server;
    3) inject nmi to guest/ make linux kernel crash;
    4) sleep long time, then reset vm via system_reset;
    5) query clock offset from ntp server;

    :param test: QEMU test object.
    :param params: Dictionary with test parameters.
    :param env: Dictionary with the test environment.
    """
    ntp_cmd = params["ntp_cmd"]
    ntp_query_cmd = params["ntp_query_cmd"]
    nmi_cmd = params.get("nmi_cmd", "inject-nmi")
    sleep_time = float(params.get("sleep_time", 1800))
    deviation = float(params.get("deviation", 5))
    os_type = params["os_type"]
    ntp_host_cmd = params.get("ntp_host_cmd", ntp_cmd)

    error_context.context("sync host time with ntp server", logging.info)
    process.system(ntp_host_cmd, shell=True)

    error_context.context("start guest", logging.info)
    params["start_vm"] = "yes"
    preprocess(test, params, env)
    vm = env.get_vm(params["main_vm"])
    vm.verify_alive()

    if params["os_type"] == 'windows':
        utils_time.sync_timezone_win(vm)

    timeout = int(params.get("login_timeout", 360))
    session = vm.wait_for_login(timeout=timeout)

    error_context.context("sync time in guest", logging.info)
    if os_type == "windows":
        utils_test.start_windows_service(session, "w32time")
    session.cmd(ntp_cmd)

    error_context.context("inject nmi interrupt in vm", logging.info)
    target, cmd = re.split(r"\s*:\s*", nmi_cmd)
    if target == "monitor":
        vm.monitor.send_args_cmd(cmd)
    else:
        session.sendline(cmd)
    try:
        session.cmd("dir")
    except Exception:
        pass
    else:
        test.fail("Guest OS still alive ...")

    error_context.context("sleep %s seconds" % sleep_time, logging.info)
    time.sleep(sleep_time)
    # Autotest parses serial output and could raise VMDeadKernelCrash
    # we generated using sysrq. Ignore one "BUG:" line
    vm.resume()
    try:
        session = vm.reboot(method="system_reset")
    except VMDeadKernelCrashError as details:
        details = str(details)
        if (re.findall(r"Trigger a crash\s.*BUG:", details, re.M) and
                details.count("BUG:") != 1):
            test.fail("Got multiple kernel crashes. Please "
                      "note that one of them was "
                      "intentionally  generated by sysrq in "
                      "this test.\n%s" % details)
        end_time = time.time() + timeout
        while time.time() < end_time:
            try:
                session = vm.wait_for_login(timeout=timeout)
            except VMDeadKernelCrashError as details:
                details = str(details)
                if (re.findall(r"Trigger a crash\s.*BUG:", details,
                               re.M) and details.count("BUG:") != 1):
                    test.fail("Got multiple kernel crashes. "
                              "Please note that one of them was "
                              "intentionally  generated by sysrq "
                              "in this test.\n%s" % details)
            else:
                break

    error_context.context("check time offset via ntp", logging.info)
    output = session.cmd_output(ntp_query_cmd)
    try:
        offset = re.findall(r"[+-]?(\d+\.\d+)", output, re.M)[-1]
    except IndexError:
        offset = 0.0
    if float(offset) > deviation:
        test.fail("Unacceptable offset '%s', " % offset +
                  "deviation '%s'" % deviation)
